/***************************************************************************
mondo-prep-freebsd.c  -  description
-----------------

begin: Fri Apr 19 16:40:35 EDT 2004
copyright : (C) 2004 Mondo  Hugo Rabson
email     : Hugo Rabson <hugorabson@msn.com>
edited by : by Stan Benoit 4/2002
email     : troff@nakedsoul.org
cvsid     : $Id: mondo-prep.c,v 1.14 2004/06/21 20:20:36 hugo Exp $
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/* mondo-prep-freebsd.c             Hugo Rabson


09/21
- removed g_fprep

08/20
- moved lots of funcs from mondo-prep-general.c to mondo-prep-linux.c
  and mondo-prep-freebsd.c

08/20/2004
- first incarnation
*/

/**
 * @file
 * Functions for prepping hard drives: partitioning, formatting, etc.
 */

#ifdef __FreeBSD__

#include "../common/my-stuff.h"
#include "../common/mondostructures.h"
#include "mondoprep.h"
#include "../common/libmondo.h"
#include "mondo-rstr-tools-EXT.h"
#include <sys/ioctl.h>
#include <linux/hdreg.h>
#include <math.h>


#define FDISK_LOG "/tmp/fdisk.log"





extern int g_partition_table_locked_up;

int set_partition_type (FILE*,const char *, int, const char *, long long);


extern void wipe_MBRs_and_reboot_if_necessary(struct mountlist_itself *mountlist);
extern int format_everything(struct mountlist_itself *mountlist, bool interactively);
extern int fput_string_one_char_at_a_time(FILE*fout, char*str);
extern void create_mountlist_for_drive(struct mountlist_itself *mountlist, char *drive_name, struct mountlist_reference *drivemntlist);
extern void resize_mountlist_proportionately_to_suit_new_drives(struct mountlist_itself
							 *mountlist);



int fput_string_one_char_at_a_time(FILE*fout, char*str)
{
  int i, j;
  FILE*fq;

  if (ferror(fout))
    {
	return(-1);
    }
  log_msg(5, "Writing string '%s', one char at a time", str);
  j = strlen(str);
  for(i=0; i<j; i++)
    {
      log_msg(6, "Writing %d ('%c')", str[i], str[i]);
      if ((fq = fopen(FDISK_LOG, "a+"))) { fputc(str[i], fq);fclose(fq); }
      fputc(str[i], fout);
      fflush(fout);
      usleep(1000L*100L);
      if (str[i]<32) { usleep(1000L*10L); }
    }
  log_msg(5, "Returning");
   
   return(i);
}










/**
 * @addtogroup prepGroup
 * @{
 */
/**
 * Execute the commands in /tmp/i-want-my-lvm.
 * These should probably be commands to set up LVM.
 * @return The number of errors encountered (0 for success).
 */
 

int do_my_funky_lvm_stuff(bool just_erase_existing_volumes, bool vacuum_pack)
{
	return (0);
}


/**
 * Add RAID partitions while copying @p old_mountlist to @p new_mountlist.
 * We go through @p old_mountlist and check if any RAID device (/dev/md? on Linux)
 * is in it; if it is, then we put the disks contained within that RAID device
 * into the mountlist as well.
 * @param old_mountlist The mountlist to read.
 * @param new_mountlist The mountlist to write, with the RAID partitions added.
 * @return 0 for success, nonzero for failure.
 */
	int extrapolate_mountlist_to_include_raid_partitions(struct mountlist_itself
						     *new_mountlist, struct mountlist_itself
						     *old_mountlist)
{
	log_to_screen ("I don't know how to extrapolate the mountlist on FreeBSD. Sorry.");
	return(1);
}


/**
 * Format @p device as a @p format filesystem.
 * This will use the format command returned by which_format_command_do_i_need().
 * If @p device is an LVM PV, it will not be formatted, and LVM will be started
 * (if not already done). If it's an imagedev, software RAID component, or
 * (under BSD) swap partition, no format will be done.
 * @param device The device to format.
 * @param format The filesystem type to format it as.
 * @return 0 for success, nonzero for failure.
 */
int format_device(char *device, char *format)
{
	/** int **************************************************************/
	int res;
	int retval = 0;
	static bool vinum_started_yet = FALSE;

	/** buffers ***********************************************************/
	char *program;
	char *tmp;

	/** end ****************************************************************/

	malloc_string(program);
	malloc_string(tmp);
	assert_string_is_neither_NULL_nor_zerolength(device);
	assert(format != NULL);

	if (strstr(format, "raid")) {	// do not form RAID disks; do it to /dev/md* instead
		sprintf(tmp, "Not formatting %s (it is a RAID disk)", device);
		log_it(tmp);
		paranoid_free(program); paranoid_free(tmp);
		return (0);
	}
	if (strcmp(format, "swap") == 0) {
		log_it("Not formatting %s - it's swap", device);
		paranoid_free(program); paranoid_free(tmp);
		return (0);
	}
	if (strlen(format) <= 2) {
		sprintf(tmp,
			"%s has a really small format type ('%s') - this is probably a hexadecimal string, which would suggest the partition is an image --- I shouldn't format it",
			device, format);
		log_it(tmp);
		paranoid_free(program); paranoid_free(tmp);
		return (0);
	}
	if (is_this_device_mounted(device)) {
		sprintf(tmp, "%s is mounted - cannot format it       ", device);
		log_to_screen(tmp);
		paranoid_free(program); paranoid_free(tmp);
		return (1);
	}
	if (strstr(device, RAID_DEVICE_STUB)) {
		newtSuspend();
	
	if (!vinum_started_yet) {
			if (!does_file_exist("/tmp/raidconf.txt")) {
				log_to_screen("/tmp/raidconf.txt does not exist. I therefore cannot start Vinum.");
			} else {
				int res;
				res = run_program_and_log_output("vinum create /tmp/raidconf.txt", TRUE);
				if (res) {
					log_to_screen
					    ("`vinum create /tmp/raidconf.txt' returned errors. Please fix them and re-run mondorestore.");
					finish(1);
				}
				vinum_started_yet = TRUE;
			}
		}

		if (vinum_started_yet) {
			FILE *fin;
			char line[MAX_STR_LEN];
			sprintf(tmp, "Initializing Vinum device %s (this may take a *long* time)", device);
			log_to_screen(tmp);
			/* format raid partition */
			//      sprintf (program, "mkraid --really-force %s", device); --- disabled -- BB, 02/12/2003
			sprintf(program,
				"for plex in `vinum lv -r %s | grep '^P' | tr '\t' ' ' | tr -s ' ' | cut -d' ' -f2`; do echo $plex; done > /tmp/plexes",
				basename(device));
			system(program);
			fin = fopen("/tmp/plexes", "r");
			while (fgets(line, MAX_STR_LEN - 1, fin)) {
				if (strchr(line, '\n'))
					*(strchr(line, '\n')) = '\0';	// get rid of the \n on the end

				sprintf(tmp, "Initializing plex: %s", line);
				open_evalcall_form(tmp);
				sprintf(tmp, "vinum init %s", line);
				system(tmp);
				while (1) {
					sprintf(tmp,
						"vinum lp -r %s | grep '^S' | head -1 | tr -s ' ' | cut -d: -f2 | cut -f1 | sed 's/^ //' | sed 's/I //' | sed 's/%%//'",
						line);
					FILE *pin = popen(tmp, "r");
					char status[MAX_STR_LEN / 4];
					fgets(status, MAX_STR_LEN / 4 - 1, pin);
					pclose(pin);

					if (!strcmp(status, "up")) {
						break;	/* it's done */
					}
					update_evalcall_form(atoi(status));
					usleep(250000);
				}
				close_evalcall_form();
			}
			fclose(fin);
			unlink("/tmp/plexes");
			/* retval+=res; */
		}
		
		system("sync"); sleep(1);
		newtResume();
	}

	if (!strcmp(format, "lvm"))
	  {
	    log_msg(1, "Don't format %s - it's part of an lvm volume", device);
	    paranoid_free(program); paranoid_free(tmp);
	    return(0);
	  }
	res = which_format_command_do_i_need(format, program);
	sprintf(tmp, "%s %s", program, device);
	if (strstr(program, "kludge")) {
		strcat(tmp, " /");
	}
	sprintf(program, "sh -c 'echo -en \"y\\ny\\ny\\n\" | %s'", tmp);
	sprintf(tmp, "Formatting %s as %s", device, format);
	update_progress_form(tmp);
	res = run_program_and_log_output(program, FALSE);
	if (res && strstr(program, "kludge")) {
		sprintf(tmp, "Kludge failed; using regular mkfs.%s to format %s", format, device);
		sprintf(program, "newfs_msdos -F 32 %s", device);
		res = run_program_and_log_output(program, FALSE);
	}
	retval += res;
	if (retval) {
		strcat(tmp, "...failed");
	} else {
		strcat(tmp, "...OK");
	}

	log_to_screen(tmp);
	paranoid_free(program); paranoid_free(tmp);
	system("sync"); sleep(1);
	return (retval);
}





/**
 * Format all drives (except those excluded by format_device()) in @p mountlist.
 * @param mountlist The mountlist containing partitions to be formatted.
 * @param interactively If TRUE, then prompt the user before each partition.
 * @return The number of errors encountered (0 for success).
 */
int format_everything(struct mountlist_itself *mountlist, bool interactively)
{
	/** int **************************************************************/
	int retval = 0;
	int lino;
	int res;
//	int i;
//	struct list_of_disks *drivelist;

	/** long *************************************************************/
	long progress_step;

	/** bools ************************************************************/
	bool do_it;

	/** buffers **********************************************************/
	char *tmp;

	/** pointers *********************************************************/
	struct mountlist_line *me;	// mountlist entry
	/** end **************************************************************/

	assert(mountlist!=NULL);
	malloc_string(tmp);
	sprintf(tmp, "format_everything (mountlist, interactively = %s", (interactively) ? "true" : "false");
	log_it(tmp);
	mvaddstr_and_log_it(g_currentY, 0, "Formatting partitions     ");
	open_progress_form("Formatting partitions", "I am now formatting your hard disk partitions.",
			   "This may take up to five minutes.", "", mountlist->entries + 1);

	progress_step = (mountlist->entries > 0) ? g_maximum_progress / mountlist->entries : 1;
// start soft-raids now (because LVM might depend on them)
// ...and for simplicity's sake, let's format them at the same time :)
	log_msg(1, "Stopping all RAID devices");
	stop_all_raid_devices(mountlist);
	system("sync"); system("sync"); system("sync"); 
	sleep(2);
	log_msg(1, "Prepare soft-RAIDs"); // prep and format too
	for (lino = 0; lino < mountlist->entries; lino++) {
		me = &mountlist->el[lino];	// the current mountlist entry
		log_msg(2, "Examining %s", me->device);
		if (!strncmp(me->device, "/dev/md", 7)) {
			if (interactively) {
				// ask user if we should format the current device
				sprintf(tmp, "Shall I format %s (%s) ?", me->device, me->mountpoint);
				do_it = ask_me_yes_or_no(tmp);
			} else {
				do_it = TRUE;
			}
			if (do_it) {
			// NB: format_device() also stops/starts RAID device if necessary
				retval += format_device(me->device, me->format);
			}
			g_current_progress += progress_step;
		}
	}
	system("sync"); system("sync"); system("sync"); 
	sleep(2);
// This last step is probably necessary
//	log_to_screen("Re-starting software RAIDs...");
//	start_all_raid_devices(mountlist);
//	system("sync"); system("sync"); system("sync"); 
//	sleep(5);
// do LVMs now
	log_msg(1, "Creating LVMs");
	if (does_file_exist("/tmp/i-want-my-lvm")) {
		wait_until_software_raids_are_prepped("/proc/mdstat", 10);
		log_to_screen("Configuring LVM");
                if (!g_text_mode) { newtSuspend(); }
/*
		for(i=0; i<3; i++)
		  {
		    res = do_my_funky_lvm_stuff(FALSE, FALSE);
		    if (!res) { break; }
		    sleep(3);
		    res = do_my_funky_lvm_stuff(TRUE, FALSE);
		    sleep(3);
		  }
		if (res) {
			log_msg(1, "Vacuum-packing...");
*/
			res = do_my_funky_lvm_stuff(FALSE, TRUE);
/*
		}
*/
                if (!g_text_mode) { newtResume(); }
		if (!res) { log_to_screen("LVM initialized OK"); }
		else { log_to_screen("Failed to initialize LVM"); }
		// retval += res;
		if (res) {
			retval ++;
		}
	    sleep(3);
	}

// do regulars at last
	sleep(2); // woo!
        log_msg(1, "Formatting regulars");
	for (lino = 0; lino < mountlist->entries; lino++) {
		me = &mountlist->el[lino];	// the current mountlist entry
		if (!strcmp(me->mountpoint, "image")) {
			sprintf(tmp, "Not formatting %s - it's an image", me->device);
			log_it(tmp);
		} else if (!strcmp(me->format, "raid")) {
			sprintf(tmp, "Not formatting %s - it's a raid-let", me->device);
			log_it(tmp);
			continue;
		} else if (!strcmp(me->format, "lvm")) {
			sprintf(tmp, "Not formatting %s - it's an LVM", me->device);
			log_it(tmp);
			continue;
		} else if (!strncmp(me->device, "/dev/md", 7)) {
			sprintf(tmp, "Already formatted %s - it's a soft-RAID dev", me->device);
			log_it(tmp);
			continue;
		} else if (!does_file_exist(me->device)
					&& strncmp(me->device, "/dev/hd", 7)
					&& strncmp(me->device, "/dev/sd", 7)) {
			sprintf(tmp, "Not formatting %s yet - doesn't exist - probably an LVM", me->device);
			log_it(tmp);
			continue;
		} else {
			if (interactively) {
				// ask user if we should format the current device
				sprintf(tmp, "Shall I format %s (%s) ?", me->device, me->mountpoint);
				do_it = ask_me_yes_or_no(tmp);
			} else {
				do_it = TRUE;
			}

			if (do_it)
				retval += format_device(me->device, me->format);
		}

		// update progress bar
		g_current_progress += progress_step;
	}


	// update progress bar to 100% to compensate for
	// rounding errors of the progress_step calculation
	if (lino >= mountlist->entries)
		g_current_progress = g_maximum_progress;

	close_progress_form();

	if (retval) {
		mvaddstr_and_log_it(g_currentY++, 74, "Failed.");
		log_to_screen("Errors occurred during the formatting of your hard drives.");
	} else {
		mvaddstr_and_log_it(g_currentY++, 74, "Done.");
	}

	sprintf(tmp, "format_everything () - %s", (retval) ? "failed!" : "finished successfully");
	log_it(tmp);

	if (g_partition_table_locked_up > 0)
	  {
	    if (retval > 0 && !interactively)
              {
//123456789 123456789 123456789 123456789 123456789 123456789 123456789 123456789 
                log_to_screen(
  "Partition table locked up %d times. At least one 'mkfs' (format) command", g_partition_table_locked_up);
                log_to_screen(
  "failed. I think these two events are related. Sometimes, fdisk's ioctl() call");
                log_to_screen(
  "to refresh its copy of the partition table causes the kernel to lock the ");
                log_to_screen(
  "partition table. I believe this has just happened.");
                if (ask_me_yes_or_no("Please choose 'yes' to reboot and try again; or 'no' to ignore this warning and continue."))
                  {
		    system("sync");
		    system("sync");
		    system("sync");
		    system("reboot"); 
		  }
	      }
	    else
	      {
	        log_to_screen(
"Partition table locked up %d time%c. However, disk formatting succeeded.", g_partition_table_locked_up, (g_partition_table_locked_up==1)?'.':'s');
	      }
	  }
        newtSuspend();
	system("clear");
	newtResume();
	paranoid_free(tmp);
	return (retval);
}


/**
 * Create small dummy partitions to fill in the gaps in partition numbering for @p drivename.
 * Each partition created is 32k in size.
 * @param drivename The drive to create the dummy partitions on.
 * @param devno_we_must_allow_for The lowest-numbered real partition; create
 * dummies up to (this - 1).
 * @return The number of errors encountered (0 for success).
 */
int make_dummy_partitions(FILE*pout_to_fdisk, char *drivename, int devno_we_must_allow_for)
{
	/** int **************************************************************/
	int current_devno;
	int previous_devno;
	int retval = 0;
	int res;

	/** buffers **********************************************************/
	char *tmp;

	/** end **************************************************************/

	malloc_string(tmp);
	assert_string_is_neither_NULL_nor_zerolength(drivename);

	if (devno_we_must_allow_for >= 5) {
		sprintf(tmp, "Making dummy primary %s%d", drivename, 1);
		log_it(tmp);
		g_maximum_progress++;
		res = partition_device(pout_to_fdisk,drivename, 1, 0, "ext2", 32000);
		retval += res;
		previous_devno = 1;
		current_devno = 5;
	} else {
		previous_devno = 0;
		current_devno = 1;
	}
	for (; current_devno < devno_we_must_allow_for; current_devno++) {
		sprintf(tmp, "Creating dummy partition %s%d", drivename, current_devno);
		log_it(tmp);
		g_maximum_progress++;
		res = partition_device(pout_to_fdisk,drivename, current_devno, previous_devno, OSSWAP("ext2", "ufs"), 32000);
		retval += res;
		previous_devno = current_devno;
	}
	paranoid_free(tmp);
	return (previous_devno);
}


/**
 * Decide whether @p mountlist contains any RAID devices.
 * @param mountlist The mountlist to examine.
 * @return TRUE if it does, FALSE if it doesn't.
 */
bool mountlist_contains_raid_devices(struct mountlist_itself * mountlist)
{
	/** int *************************************************************/
	int i;
	int matching = 0;

	/** end **************************************************************/

	assert(mountlist != NULL);

	for (i = 0; i < mountlist->entries; i++) {
		if (strstr(mountlist->el[i].device, RAID_DEVICE_STUB)) {
			matching++;
		}
	}
	if (matching) {
		return (TRUE);
	} else {
		return (FALSE);
	}
}

/* The following 2 functions are stolen from /usr/src/sbin/disklabel/disklabel.c */
static void display_disklabel(FILE * f, const struct disklabel *lp)
{
	int i, j;
	const struct partition *pp;

	fprintf(f, "# %s\n", "Generated by Mondo Rescue");
	if (lp->d_type < DKMAXTYPES)
		fprintf(f, "type: %s\n", dktypenames[lp->d_type]);
	else
		fprintf(f, "type: %u\n", lp->d_type);
	fprintf(f, "disk: %.*s\n", (int) sizeof(lp->d_typename), lp->d_typename);
	fprintf(f, "label: %.*s\n", (int) sizeof(lp->d_packname), lp->d_packname);
	fprintf(f, "flags:");
	if (lp->d_flags & D_REMOVABLE)
		fprintf(f, " removeable");
	if (lp->d_flags & D_ECC)
		fprintf(f, " ecc");
	if (lp->d_flags & D_BADSECT)
		fprintf(f, " badsect");
	fprintf(f, "\n");
	fprintf(f, "bytes/sector: %lu\n", (u_long) lp->d_secsize);
	fprintf(f, "sectors/track: %lu\n", (u_long) lp->d_nsectors);
	fprintf(f, "tracks/cylinder: %lu\n", (u_long) lp->d_ntracks);
	fprintf(f, "sectors/cylinder: %lu\n", (u_long) lp->d_secpercyl);
	fprintf(f, "cylinders: %lu\n", (u_long) lp->d_ncylinders);
	fprintf(f, "sectors/unit: %lu\n", (u_long) lp->d_secperunit);
	fprintf(f, "rpm: %u\n", lp->d_rpm);
	fprintf(f, "interleave: %u\n", lp->d_interleave);
	fprintf(f, "trackskew: %u\n", lp->d_trackskew);
	fprintf(f, "cylinderskew: %u\n", lp->d_cylskew);
	fprintf(f, "headswitch: %lu\t\t# milliseconds\n", (u_long) lp->d_headswitch);
	fprintf(f, "track-to-track seek: %ld\t# milliseconds\n", (u_long) lp->d_trkseek);
	fprintf(f, "drivedata: ");
	for (i = NDDATA - 1; i >= 0; i--)
		if (lp->d_drivedata[i])
			break;
	if (i < 0)
		i = 0;
	for (j = 0; j <= i; j++)
		fprintf(f, "%lu ", (u_long) lp->d_drivedata[j]);
	fprintf(f, "\n\n%u partitions:\n", lp->d_npartitions);
	fprintf(f, "#        size   offset    fstype   [fsize bsize bps/cpg]\n");
	pp = lp->d_partitions;
	for (i = 0; i < lp->d_npartitions; i++, pp++) {
		if (pp->p_size) {
			fprintf(f, "  %c: %8lu %8lu  ", 'a' + i, (u_long) pp->p_size, (u_long) pp->p_offset);
			if (pp->p_fstype < FSMAXTYPES)
				fprintf(f, "%8.8s", fstypenames[pp->p_fstype]);
			else
				fprintf(f, "%8d", pp->p_fstype);
			switch (pp->p_fstype) {

			case FS_UNUSED:	/* XXX */
				fprintf(f, "    %5lu %5lu %5.5s ", (u_long) pp->p_fsize, (u_long) (pp->p_fsize * pp->p_frag), "");
				break;

			case FS_BSDFFS:
				fprintf(f, "    %5lu %5lu %5u ", (u_long) pp->p_fsize, (u_long) (pp->p_fsize * pp->p_frag),
					pp->p_cpg);
				break;

			case FS_BSDLFS:
				fprintf(f, "    %5lu %5lu %5d", (u_long) pp->p_fsize, (u_long) (pp->p_fsize * pp->p_frag),
					pp->p_cpg);
				break;

			default:
				fprintf(f, "%20.20s", "");
				break;
			}
			fprintf(f, "\t# (Cyl. %4lu", (u_long) (pp->p_offset / lp->d_secpercyl));
			if (pp->p_offset % lp->d_secpercyl)
				putc('*', f);
			else
				putc(' ', f);
			fprintf(f, "- %lu", (u_long) ((pp->p_offset + pp->p_size + lp->d_secpercyl - 1) / lp->d_secpercyl - 1));
			if (pp->p_size % lp->d_secpercyl)
				putc('*', f);
			fprintf(f, ")\n");
		}
	}
	fflush(f);
}

static struct disklabel *get_virgin_disklabel(char *dkname)
{
	static struct disklabel loclab;
	struct partition *dp;
	char lnamebuf[BBSIZE];
	int f;
	u_int secsize, u;
	off_t mediasize;

	(void) snprintf(lnamebuf, BBSIZE, "%s", dkname);
	if ((f = open(lnamebuf, O_RDONLY)) == -1) {
		warn("cannot open %s", lnamebuf);
		return (NULL);
	}

	/* New world order */
	if ((ioctl(f, DIOCGMEDIASIZE, &mediasize) != 0)
	    || (ioctl(f, DIOCGSECTORSIZE, &secsize) != 0)) {
		close(f);
		return (NULL);
	}
	memset(&loclab, 0, sizeof loclab);
	loclab.d_magic = DISKMAGIC;
	loclab.d_magic2 = DISKMAGIC;
	loclab.d_secsize = secsize;
	loclab.d_secperunit = mediasize / secsize;

	/*
	 * Nobody in these enligthened days uses the CHS geometry for
	 * anything, but nontheless try to get it right.  If we fail
	 * to get any good ideas from the device, construct something
	 * which is IBM-PC friendly.
	 */
	if (ioctl(f, DIOCGFWSECTORS, &u) == 0)
		loclab.d_nsectors = u;
	else
		loclab.d_nsectors = 63;
	if (ioctl(f, DIOCGFWHEADS, &u) == 0)
		loclab.d_ntracks = u;
	else if (loclab.d_secperunit <= 63 * 1 * 1024)
		loclab.d_ntracks = 1;
	else if (loclab.d_secperunit <= 63 * 16 * 1024)
		loclab.d_ntracks = 16;
	else
		loclab.d_ntracks = 255;
	loclab.d_secpercyl = loclab.d_ntracks * loclab.d_nsectors;
	loclab.d_ncylinders = loclab.d_secperunit / loclab.d_secpercyl;
	loclab.d_npartitions = MAXPARTITIONS;

	/* Various (unneeded) compat stuff */
	loclab.d_rpm = 3600;
	loclab.d_bbsize = BBSIZE;
	loclab.d_interleave = 1;;
	strncpy(loclab.d_typename, "amnesiac", sizeof(loclab.d_typename));

	dp = &loclab.d_partitions[RAW_PART];
	dp->p_size = loclab.d_secperunit;
	loclab.d_checksum = dkcksum(&loclab);
	close(f);
	return (&loclab);
}
/* End stolen from /usr/src/sbin/disklabel/disklabel.c. */

char *canonical_name(char *drivename)
{
	if (drivename) {
		if (strncmp(drivename, "/dev/", 5) == 0) {
			return drivename + 5;
		}
	}
	return drivename;
}

/**
 * (BSD only) Create a disklabel on @p drivename according to @p mountlist.
 * @param mountlist The mountlist to get the subpartition information from.
 * @param drivename The drive or slice to create a disklabel on.
 * @param ret If non-NULL, store the created disklabel here.
 * @return The number of errors encountered (0 for success).
 */
int label_drive_or_slice(struct mountlist_itself *mountlist, char *drivename, struct disklabel *ret)
{
	char subdev_str[MAX_STR_LEN];
	char command[MAX_STR_LEN];
	struct disklabel *lp;
	int i, lo = 0;
	int retval = 0;
	char c;
	FILE *ftmp;

	lp = get_virgin_disklabel(drivename);
	for (c = 'a'; c <= 'z'; ++c) {
		int idx;
		sprintf(subdev_str, "%s%c", drivename, c);
		if ((idx = find_device_in_mountlist(mountlist, subdev_str)) < 0) {
			lp->d_partitions[c - 'a'].p_size = 0;
			lp->d_partitions[c - 'a'].p_fstype = FS_UNUSED;
		} else {
			lo = c - 'a';
			lp->d_partitions[c - 'a'].p_size = mountlist->el[idx].size * 2;
			lp->d_partitions[c - 'a'].p_fsize = 0;
			lp->d_partitions[c - 'a'].p_frag = 0;
			lp->d_partitions[c - 'a'].p_cpg = 0;
			if (!strcmp(mountlist->el[idx].format, "ufs")
			    || !strcmp(mountlist->el[idx].format, "ffs")
			    || !strcmp(mountlist->el[idx].format, "4.2BSD")) {
				lp->d_partitions[c - 'a'].p_fstype = FS_BSDFFS;
				lp->d_partitions[c - 'a'].p_fsize = 2048;
				lp->d_partitions[c - 'a'].p_frag = 8;
				lp->d_partitions[c - 'a'].p_cpg = 64;
			} else if (!strcasecmp(mountlist->el[idx].format, "raid")
				   || !strcasecmp(mountlist->el[idx].format, "vinum")) {
				lp->d_partitions[c - 'a'].p_fstype = FS_VINUM;
			} else if (!strcmp(mountlist->el[idx].format, "swap")) {
				lp->d_partitions[c - 'a'].p_fstype = FS_SWAP;
			} else
				lp->d_partitions[c - 'a'].p_fstype = FS_OTHER;
		}
	}

	// fix up the offsets
	lp->d_partitions[0].p_offset = 0;
	lp->d_partitions[RAW_PART].p_offset = 0;
	lp->d_partitions[RAW_PART].p_size = lp->d_secperunit;
	lp->d_partitions[RAW_PART].p_fstype = FS_UNUSED;

	for (i = 1; i < lp->d_npartitions; ++i) {
		int lastone;
		if ((i == RAW_PART) || (lp->d_partitions[i].p_size == 0))
			continue;
		for (lastone = i - 1; lastone >= 0; lastone--) {
			if ((lp->d_partitions[lastone].p_size)
			    && (lastone != RAW_PART))
				break;
		}
		lp->d_partitions[i].p_offset = lp->d_partitions[lastone].p_offset + lp->d_partitions[lastone].p_size;
	}
	if (lp->d_partitions[lo].p_offset + lp->d_partitions[lo].p_size > lp->d_secperunit) {
		lp->d_partitions[lo].p_size = lp->d_secperunit - lp->d_partitions[lo].p_offset;
	}

	ftmp = fopen("/tmp/disklabel", "w");
	display_disklabel(ftmp, lp);
	fclose(ftmp);
	sprintf(command, "disklabel -wr %s auto", canonical_name(drivename));
	retval += run_program_and_log_output(command, TRUE);
	sprintf(command, "disklabel -R %s /tmp/disklabel", canonical_name(drivename));
	retval += run_program_and_log_output(command, TRUE);
	if (ret)
		*ret = *lp;
	return retval;
}


/**
 * Partition @p drivename based on @p mountlist.
 * @param mountlist The mountlist to use to guide the partitioning.
 * @param drivename The drive to partition.
 * @return 0 for success, nonzero for failure.
 */
int partition_drive(struct mountlist_itself *mountlist, char *drivename)
{
	/** int *************************************************************/
	int current_devno;
	int previous_devno = 0;
	int lino;
	int retval = 0;
	int i;
	FILE*pout_to_fdisk=NULL;
	
	bool fbsd_part = FALSE;
	char subdev_str[MAX_STR_LEN];

	/** long long *******************************************************/
	long long partsize;

	/** buffers *********************************************************/
	char *device_str;
	char *format;
	char *tmp;

	/** end *************************************************************/

	assert(mountlist != NULL);
	assert_string_is_neither_NULL_nor_zerolength(drivename);

	malloc_string(device_str);
	malloc_string(format);
	malloc_string(tmp);
	
	sprintf(tmp, "Partitioning drive %s", drivename);
	log_it(tmp);

        log_it("(Not opening fdisk now; that's the Linux guy's job)");
	pout_to_fdisk = NULL;
	sprintf(tmp, "fdisk %s >> %s 2>> %s", drivename, FDISK_LOG, FDISK_LOG);
	pout_to_fdisk = popen(tmp, "w");
	if (!pout_to_fdisk) { 
		log_to_screen("Cannot call fdisk to configure %s", drivename); 
		paranoid_free(device_str);
		paranoid_free(format);
		paranoid_free(tmp);
		return(1); 
	}
	for (current_devno = 1; current_devno < 99; current_devno++) {
		build_partition_name(device_str, drivename, current_devno);
		lino = find_device_in_mountlist(mountlist, device_str);

		if (lino < 0) {
			// device not found in mountlist
			// If this is the first partition (just as a sentinel value),
			// then see if the user has picked 'dangerously-dedicated' mode.
			// If so, then we just call label_drive_or_slice() and return.
			char c;
			if (current_devno == 1) {
				// try DangerouslyDedicated mode
				for (c = 'a'; c <= 'z'; c++) {
					sprintf(subdev_str, "%s%c", drivename, c);
					if (find_device_in_mountlist(mountlist, subdev_str) > 0) {
						fbsd_part = TRUE;
					}
				}
				if (fbsd_part) {
					int r = label_drive_or_slice(mountlist,
								     drivename,
								     0);
					char command[MAX_STR_LEN];
					sprintf(command, "disklabel -B %s", basename(drivename));
					if (system(command)) {
						log_to_screen("Warning! Unable to make the drive bootable.");
					}
					paranoid_free(device_str);
					paranoid_free(format);
					paranoid_free(tmp);
					return r;
				}
			}
			for (c = 'a'; c <= 'z'; c++) {
				sprintf(subdev_str, "%s%c", device_str, c);
				if (find_device_in_mountlist(mountlist, subdev_str) > 0) {
					fbsd_part = TRUE;
				}
			}
			// Now we check the subpartitions of the current partition.
			if (fbsd_part) {
				int i, line;

				strcpy(format, "ufs");
				partsize = 0;
				for (i = 'a'; i < 'z'; ++i) {
					sprintf(subdev_str, "%s%c", device_str, i);
					line = find_device_in_mountlist(mountlist, subdev_str);
					if (line > 0) {
						// We found one! Add its size to the total size.
						partsize += mountlist->el[line].size;
					}
				}
			} else {
				continue;
			}
		}

		/* OK, we've found partition /dev/hdxN in mountlist; let's prep it */
		/* For FreeBSD, that is      /dev/adXsY */

		log_it("Found partition %s in mountlist", device_str);
		if (!previous_devno) {

			log_it("Wiping %s's partition table", drivename);
			// FreeBSD doesn't let you write to blk devices in <512byte chunks.
//			sprintf(tmp, "dd if=/dev/zero of=%s count=1 bs=512", drivename);
//			if (run_program_and_log_output(tmp, TRUE)) {
			file = open(drivename, O_WRONLY);
			if (!file) {
				sprintf(tmp, "Warning - unable to open %s for wiping it's partition table", drivename);
				log_to_screen(tmp);
			}

			for (i=0; i<512; i++) {
				if ( !write(file, "\0", 1) ) {
					sprintf(tmp, "Warning - unable to write to %s", drivename);
					log_to_screen(tmp);
				}
			}
			system("sync");
			if (current_devno > 1) {
				previous_devno = make_dummy_partitions(pout_to_fdisk,drivename, current_devno);
			}
		}
		if (!fbsd_part) {

		strcpy(format, mountlist->el[lino].format);
		partsize = mountlist->el[lino].size;

		}

		if (current_devno == 5 && previous_devno == 4) {
			log_to_screen("You must leave at least one partition spare as the Extended partition.");
			paranoid_free(device_str);
			paranoid_free(format);
			paranoid_free(tmp);
			return (1);
		}

		retval += partition_device(pout_to_fdisk, drivename, current_devno, previous_devno, format, partsize);

		if ((current_devno <= 4) && fbsd_part) {
			sprintf(tmp, "disklabel -B %s", basename(device_str));
			retval += label_drive_or_slice(mountlist, device_str, 0);
			if (system(tmp)) {
				log_to_screen("Warning! Unable to make the slice bootable.");
			}
		}

		previous_devno = current_devno;
	}

	if (pout_to_fdisk)
	{
// mark relevant partition as bootable
		sprintf(tmp, "a\n%s\n", call_program_and_get_last_line_of_output("make-me-bootable /tmp/mountlist.txt dummy"));
		fput_string_one_char_at_a_time(pout_to_fdisk, tmp);
// close fdisk
		fput_string_one_char_at_a_time(pout_to_fdisk, "w\n");
		system("sync");
		paranoid_pclose(pout_to_fdisk); 
		log_msg(0, "------------------- fdisk.log looks like this ------------------");
		sprintf(tmp, "cat %s >> %s", FDISK_LOG, MONDO_LOGFILE);
		system(tmp);
		log_msg(0, "------------------- end of fdisk.log... word! ------------------");
		sprintf(tmp, "tail -n6 %s | fgrep \"16: \"", FDISK_LOG);
		if (!run_program_and_log_output(tmp, 5))
		  {
		    g_partition_table_locked_up++;
		    log_to_screen("A flaw in the Linux kernel has locked the partition table.");
		  }
	}
	paranoid_free(device_str);
	paranoid_free(format);
	paranoid_free(tmp);
	return (retval);
}

/**
 * Create partition number @p partno on @p drive with @p fdisk.
 * @param drive The drive to create the partition on.
//  * @param partno The partition number of the new partition (1-4 are primary, >=5 is logical).
 * @param prev_partno The partition number of the most recently prepped partition.
 * @param format The filesystem type of this partition (used to set the type).
 * @param partsize The size of the partition in @e bytes.
 * @return 0 for success, nonzero for failure.
 */
int partition_device(FILE*pout_to_fdisk, const char *drive, int partno, int prev_partno, const char *format, long long partsize)
{
	/** int **************************************************************/
	int retval = 0;
	int res = 0;

	/** buffers **********************************************************/
	char *program;
	char *partition_name;
	char *tmp;
	char *logfile;
	char *output;

	/** pointers **********************************************************/
	char *p;
	FILE *fout;

	/** end ***************************************************************/

	malloc_string(program);
	malloc_string(partition_name);
	malloc_string(tmp);
	malloc_string(logfile);
	malloc_string(output);
	
	assert_string_is_neither_NULL_nor_zerolength(drive);
	assert(format != NULL);

	log_it("partition_device('%s', %d, %d, '%s', %lld) --- starting", drive, partno, prev_partno, format, partsize);

	if (!strncmp(drive, RAID_DEVICE_STUB, strlen(RAID_DEVICE_STUB))) {
		sprintf(tmp, "Not partitioning %s - it is a virtual drive", drive);
		log_it(tmp);
		paranoid_free(program);
		paranoid_free(partition_name);
		paranoid_free(tmp);
		paranoid_free(logfile);
		paranoid_free(output);
		return (0);
	}
	build_partition_name(partition_name, drive, partno);
	if (partsize <= 0) {
		sprintf(tmp, "Partitioning device %s (max size)", partition_name);
	} else {
		sprintf(tmp, "Partitioning device %s (%lld MB)", partition_name, (long long) partsize / 1024);
	}
	update_progress_form(tmp);
	log_it(tmp);

	if (is_this_device_mounted(partition_name)) {
		sprintf(tmp, "%s is mounted, and should not be partitioned", partition_name);
		log_to_screen(tmp);
		paranoid_free(program);
		paranoid_free(partition_name);
		paranoid_free(tmp);
		paranoid_free(logfile);
		paranoid_free(output);
		return (1);
/*
	} else if (does_partition_exist(drive, partno)) {
		sprintf(tmp, "%s already has a partition", partition_name);
		log_to_screen(tmp);
		return (1);
*/
	}


	/*  sprintf(tmp,"Partitioning %s  ",partition_name); */
	/*  mvaddstr_and_log_it(g_currentY+1,30,tmp); */
	p = (char *) strrchr(partition_name, '/');
	sprintf(logfile, "/tmp/fdisk.log.%s", ++p);
	sprintf(program, "parted2fdisk %s >> %s 2>> %s", drive, MONDO_LOGFILE, MONDO_LOGFILE);

	output[0] = '\0';
	/* make it a primary/extended/logical */
	if (partno <= 4) {
		sprintf(output + strlen(output), "n\np\n%d\n", partno);
	} else {
		if (partno == 5) {
			if (prev_partno >= 4) {
				log_to_screen("You need to leave at least one partition free, for 'extended/logical'");
				paranoid_free(program);
				paranoid_free(partition_name);
				paranoid_free(tmp);
				paranoid_free(logfile);
				paranoid_free(output);
				return (1);
			} else {
				sprintf(output + strlen(output), "n\ne\n%d\n\n\n", prev_partno + 1);
			}
		}
		strcat(output + strlen(output), "n\nl\n");
	}
	strcat(output + strlen(output), "\n");	/*start block (ENTER for next free blk */
	if (partsize > 0) {
                if (!strcmp(format, "7")) { log_msg(1, "Adding 512K, just in case"); partsize+=512; }
		sprintf(output + strlen(output), "+%lldK", (long long) (partsize));
	}
	strcat(output + strlen(output), "\n");
#if 0
/*
#endif
	sprintf(tmp,"PARTSIZE = +%ld",(long)partsize);
	log_it(tmp);
	log_it("---fdisk command---");
	log_it(output);
	log_it("---end of fdisk---");
#if 0
*/
#endif


	if (pout_to_fdisk)
	{
		log_msg(1, "Doing the new all-in-one fdisk thing");
		log_msg(1, "output = '%s'", output);
		fput_string_one_char_at_a_time(pout_to_fdisk, output);
		fput_string_one_char_at_a_time(pout_to_fdisk, "\n\np\n");
		strcpy(tmp, last_line_of_file(FDISK_LOG));
		if (strstr(tmp, " (m "))
		{
			log_msg(1, "Successfully created %s%d", drive, partno);
		}
		else
		{
			log_msg(1, "last line = %s", tmp);
			log_msg(1, "Failed to create %s%d; sending 'Enter'...", drive, partno);
		}
		if (!retval)
		{
			log_msg(1, "Trying to set %s%d's partition type now", drive, partno);
			retval = set_partition_type(pout_to_fdisk, drive, partno, format, partsize);
			if (retval)
			{
				log_msg(1, "Failed. Trying again...");
		    		retval = set_partition_type(pout_to_fdisk, drive, partno, format, partsize);
			}
		  }
		if (retval) { log_msg(1, "...but failed to set type"); }
	}
	else
	{
		strcat(output, "w\n\n");
		/* write to disk; close fdisk's stream */
		if (!(fout = popen(program, "w"))) {
			log_OS_error("can't popen-out to program");
		} else {
			fputs(output, fout);
			paranoid_pclose(fout);
		}
		if (!does_partition_exist(drive, partno) && partsize > 0) {
			log_it("Vaccum-packing");
			g_current_progress--;
			res = partition_device(pout_to_fdisk, drive, partno, prev_partno, format, -1);
			if (res) {
				sprintf(tmp, "Failed to vacuum-pack %s", partition_name);
				log_it(tmp);
				retval++;
			} else {
				retval = 0;
			}
		}
		if (does_partition_exist(drive, partno)) {
			retval = set_partition_type(pout_to_fdisk, drive, partno, format, partsize);
			if (retval) {
				sprintf(tmp, "Partitioned %s but failed to set its type", partition_name);
				log_it(tmp);
			} else {
				if (partsize > 0) {
					sprintf(tmp, "Partition %s created+configured OK", partition_name);
					log_to_screen(tmp);
				} else {
					log_it("Returning from a successful vacuum-pack");
				}
			}
			} else {
			sprintf(tmp, "Failed to partition %s", partition_name);
			if (partsize > 0) {
				log_to_screen(tmp);
			} else {
				log_it(tmp);
			}
			retval++;
		}
	}
	g_current_progress++;
	log_it("partition_device() --- leaving");
	paranoid_free(program);
	paranoid_free(partition_name);
	paranoid_free(tmp);
	paranoid_free(logfile);
	paranoid_free(output);
	return (retval);
}



/**
 * Create all partitions listed in @p mountlist.
 * @param mountlist The mountlist to use to guide the partitioning.
 * @return The number of errors encountered (0 for success).
 * @note This sets the partition types but doesn't actually do the formatting.
 * Use format_everything() for that.
 */
int partition_everything(struct mountlist_itself *mountlist)
{
	/** int ************************************************************/
	int lino;
	int retval = 0;
	int i;
	int res;

	/** buffer *********************************************************/
	struct list_of_disks *drivelist;
	/*  struct mountlist_itself new_mtlist, *mountlist; */

	/** end ************************************************************/

	drivelist = malloc(sizeof(struct list_of_disks));
	assert(mountlist != NULL);

	log_it("partition_everything() --- starting");
	mvaddstr_and_log_it(g_currentY, 0, "Partitioning hard drives      ");
	/*  mountlist=orig_mtlist; */
	if (mountlist_contains_raid_devices(mountlist)) {
		/*      mountlist=&new_mtlist; */
		/*      extrapolate_mountlist_to_include_raid_partitions(mountlist,orig_mtlist); */
		log_msg(0,"Mountlist, including the partitions incorporated in RAID devices:-");
		for (i = 0; i < mountlist->entries; i++) {
			log_it(mountlist->el[i].device);
		}
		log_msg(0, "End of mountlist.");
	}
	log_msg(0, "Stopping all LVMs, just in case");
        if (!g_text_mode) { newtSuspend(); }
	do_my_funky_lvm_stuff(TRUE, FALSE); // just remove old partitions
        if (!g_text_mode) { newtResume(); }
	log_msg(0, "Stopping all software RAID devices, just in case");
	stop_all_raid_devices(mountlist);
	log_msg(0, "Done.");
	
/*	
	if (does_file_exist("/tmp/i-want-my-lvm"))
	  {
	    wipe_MBRs_and_reboot_if_necessary(mountlist); // i.e. if it wasn't done recently
	  }
*/

	open_progress_form("Partitioning devices", "I am now going to partition all your drives.",
			   "This should not take more than five minutes.", "", mountlist->entries);

	make_list_of_drives_in_mountlist(mountlist, drivelist);

	/* partition each drive */
	for (lino = 0; lino < drivelist->entries; lino++) {
		res = partition_drive(mountlist, drivelist->el[lino].device);
		retval += res;
	}
	close_progress_form();
	if (retval) {
		mvaddstr_and_log_it(g_currentY++, 74, "Failed.");
		log_to_screen("Errors occurred during the partitioning of your hard drives.");
	} else {
		mvaddstr_and_log_it(g_currentY++, 74, "Done.");
		paranoid_system("rm -f /tmp/fdisk*.log 2> /dev/null");
	}
	newtSuspend();
	system("clear");
	newtResume();
	paranoid_free(drivelist);
	return (retval);
}






/**
 * Set the type of partition number @p partno on @p drive to @p format.
 * @param drive The drive to change the type of a partition on.
 * @param partno The partition number on @p drive to change the type of.
 * @param format The filesystem type this partition will eventually contain.
 * @param partsize The size of this partition, in @e bytes (used for vfat
 * type calculations).
 * @return 0 for success, nonzero for failure.
 */
int set_partition_type(FILE *pout_to_fdisk, const char *drive, int partno, const char *format, long long partsize)
{
	/** buffers *********************************************************/
  char *partition;
  char *command;
  char *output;
  char *tmp;
  char *partcode;
  char *logfile;

	/** pointers *********************************************************/
	char *p;
	FILE *fout;

	/** int **************************************************************/
	int res = 0;

	/** end **************************************************************/

	assert_string_is_neither_NULL_nor_zerolength(drive);
	assert(format != NULL);

  malloc_string(partition);
  malloc_string(command);
  malloc_string(output);
  malloc_string(tmp);
  malloc_string(partcode);
  malloc_string(logfile);

	build_partition_name(partition, drive, partno);
	p = (char *) strrchr(partition, '/');
	sprintf(logfile, "/tmp/fdisk-set-type.%s.log", ++p);
	if (strcmp(format, "swap") == 0) {
		strcpy(partcode, "82");
	} else if (strcmp(format, "vfat") == 0) {
		if (partsize / 1024 > 8192) {
			strcpy(partcode, "c");
		} else {
			strcpy(partcode, "b");
		}
	} else if (strcmp(format, "ext2") == 0 || strcmp(format, "reiserfs") == 0 || strcmp(format, "ext3") == 0
		   || strcmp(format, "xfs") == 0 || strcmp(format, "jfs") == 0) {
		strcpy(partcode, "83");
	} else if (strcmp(format, "minix") == 0) {
		strcpy(partcode, "81");
	} else if (strcmp(format, "raid") == 0) {
		strcpy(partcode, "fd");
	} else if ((strcmp(format, "ufs") == 0)
		   || (strcmp(format, "ffs") == 0)) {	/* raid autodetect */
		strcpy(partcode, "a5");
	} else if (strcmp(format, "lvm") == 0) {
		strcpy(partcode, "8e");
	} else if (format[0] == '\0') {	/* LVM physical partition */
		partcode[0] = '\0';
	} else if (strlen(format) >= 1 && strlen(format) <= 2) {
		strcpy(partcode, format);
	} else {
		/* probably an image */
		sprintf(tmp, "Unknown format ('%s') - using supplied string anyway", format);
		mvaddstr_and_log_it(g_currentY++, 0, tmp);
		strcpy(partcode, format); // was a5
	}
	sprintf(tmp, "Setting %s's type to %s (%s)", partition, format, partcode);
	log_msg(1, tmp);
	if (partcode[0] != '\0' && strcmp(partcode, "83")) {	/* no need to set type if 83: 83 is default */
		
		if (pout_to_fdisk)
		{
			res = 0;
			fput_string_one_char_at_a_time(pout_to_fdisk, "t\n");
			if (partno > 1 || strstr(last_line_of_file(FDISK_LOG), " (1-4)"))
			{
				log_msg(5, "Specifying partno (%d) - yay", partno);
				sprintf(tmp, "%d\n", partno);
				fput_string_one_char_at_a_time(pout_to_fdisk, tmp);
				log_msg(5, "A - last line = '%s'", last_line_of_file(FDISK_LOG));
			}
			
			sprintf(tmp, "%s\n", partcode);
			fput_string_one_char_at_a_time(pout_to_fdisk, tmp);
			log_msg(5, "B - last line = '%s'", last_line_of_file(FDISK_LOG));
			fput_string_one_char_at_a_time(pout_to_fdisk, "\n");
			log_msg(5, "C - last line = '%s'", last_line_of_file(FDISK_LOG));
			
			strcpy(tmp, last_line_of_file(FDISK_LOG));
			if (!strstr(tmp, " (m "))
			  { 
			    log_msg(1, "last line = '%s'; part type set failed", tmp);
			    res++; 
			    fput_string_one_char_at_a_time(pout_to_fdisk,"\n");
			  }
			fput_string_one_char_at_a_time(pout_to_fdisk,"p\n");
		}
		else
		{
			sprintf(output, "t\n%d\n%s\n", partno, partcode);
			strcat(output, "w\n");
			sprintf(command, "parted2fdisk %s >> %s 2>> %s", drive, MONDO_LOGFILE, MONDO_LOGFILE);
			log_msg(5, "output = '%s'", output);
			log_msg(5, "partno=%d; partcode=%s", partno, partcode);
			log_msg(5, "command = '%s'", command);
			fout = popen(command, "w");
			if (!fout) {
				log_OS_error(command);
				res = 1;
			} else {
				res = 0;
				fprintf(fout, output);
				paranoid_pclose(fout);
			}
		}
	if (res)
	  {
	    log_OS_error(command);
	  }	
	}

  paranoid_free(partition);
  paranoid_free(command);
  paranoid_free(output);
  paranoid_free(tmp);
  paranoid_free(partcode);
  paranoid_free(logfile);

  return (res);
}


int start_raid_device(char *raid_device)
{
	/** int *************************************************************/
	int res;
	int retval = 0;

	/** buffers *********************************************************/
	char *program;

	/** end *************************************************************/

	assert_string_is_neither_NULL_nor_zerolength(raid_device);
	malloc_string(program);
	
	if (is_this_device_mounted(raid_device)) {
		log_it("Can't start %s when it's mounted!", raid_device);
		return 1;
	}
	sprintf(program, "vinum start -f %s", raid_device);
	log_msg(1, "program = %s", program);
	res = run_program_and_log_output(program, 1);
	if (res) { log_msg(1, "Warning - failed to start RAID device %s", raid_device); }
	retval += res;
	sleep(1);
	return (retval);
}



/**
 * Stop @p raid_device using @p raidstop.
 * @param raid_device The software RAID device to stop.
 * @return 0 for success, nonzero for failure.
 */
int stop_raid_device(char *raid_device)
{
	/** int *************************************************************/
	int res;
	int retval = 0;

	/** buffers *********************************************************/
	char *program;

	/** end *************************************************************/

	assert_string_is_neither_NULL_nor_zerolength(raid_device);
	malloc_string(program);
	if (is_this_device_mounted(raid_device)) {
		log_it("Can't stop %s when it's mounted!", raid_device);
		return 1;
	}
	sprintf(program, "vinum stop -f %s", raid_device);
	log_msg(1, "program = %s", program);
	res = run_program_and_log_output(program, 1);
	if (res) { log_msg(1, "Warning - failed to stop RAID device %s", raid_device); }
	retval += res;
	return (retval);
}


int start_all_raid_devices(struct mountlist_itself *mountlist)
{
	int i;
	int retval=0;
	int res;
	
	for(i=0; i<mountlist->entries; i++)
	  {
	    if (!strncmp(mountlist->el[i].device, RAID_DEVICE_STUB, strlen(RAID_DEVICE_STUB)))
	      {
	        log_msg(1, "Starting %s", mountlist->el[i].device);
		res = start_raid_device(mountlist->el[i].device);
		retval += res;
	      }
	  }
	if (retval) { log_msg(1, "Started all s/w raid devices OK"); }
	else { log_msg(1, "Failed to start some/all s/w raid devices"); }
	return(retval);
}

/**
 * Stop all software RAID devices listed in @p mountlist.
 * @param mountlist The mountlist to stop the RAID devices in.
 * @return The number of errors encountered (0 for success).
 * @bug @p mountlist is not used.
 */
int stop_all_raid_devices(struct mountlist_itself *mountlist)
{
	/** int *************************************************************/
	int retval = 0;

	/** char ************************************************************/
	char *incoming;
	/** pointers ********************************************************/
	FILE *fin;
	int i;

	/** end ****************************************************************/

	malloc_string(dev);
	malloc_string(incoming);
	assert(mountlist != NULL);

	for (i = 0; i < 3; i++) {
		fin = popen("vinum list | grep '^[PVS]' | sed 's/S/1/;s/P/2/;s/V/3/' | sort | cut -d' ' -f2", "r");
		if (!fin) {
			paranoid_free(dev);
			paranoid_free(incoming);
			return (1);
		}
		for (fgets(incoming, MAX_STR_LEN - 1, fin); !feof(fin); fgets(incoming, MAX_STR_LEN - 1, fin)) {
			retval += stop_raid_device(incoming);
		}
	}
	paranoid_fclose(fin);
	if (retval) { log_msg(1, "Warning - unable to stop some RAID devices"); }
	paranoid_free(dev);
	paranoid_free(incoming);
	system("sync"); system("sync"); system("sync"); 
	sleep(1);
	return (retval);
}



/**
 * Decide which command we need to use to format a device of type @p format.
 * @param format The filesystem type we are about to format.
 * @param program Where to put the binary name for this format.
 * @return 0 for success, nonzero for failure.
 */
int which_format_command_do_i_need(char *format, char *program)
{
	/** int *************************************************************/
	int res = 0;

	/** buffers *********************************************************/
	char *tmp;

	/** end ***************************************************************/

	malloc_string(tmp);
	assert_string_is_neither_NULL_nor_zerolength(format);
	assert(program != NULL);

	if (strcmp(format, "swap") == 0) {
		strcpy(program, "true");
	} else if (strcmp(format, "vfat") == 0) {
		strcpy(program, "format-and-kludge-vfat");
	} else if (strcmp(format, "ext2") == 0) {
		strcpy(program, "mke2fs -F -q");
	} else {
		sprintf(program, "newfs_%s", format);
		sprintf(tmp, "Unknown format (%s) - assuming '%s' will do", format, program);
		log_it(tmp);
		res = 0;
	}
	paranoid_free(tmp);
	return (res);
}



/* @} - end of prepGroup */
#endif
